// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of paramxx
//
// paramxx is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// paramxx is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <memory>

#include <stdio.h>
#include <cstdlib>
#include <cstring>

using namespace std;

#include "Par.h"
#include <suplib/str.h>

Par::Par( ParTxt& par ) {

  char** my_argv = (char**) par;

  if ( my_argv ) {

    // A boolean, integer, real or string parameter
    int ii = 0;
    while( my_argv[ ii ] )
      parameter.push_back( my_argv[ ii++ ] );

  } else {

    // A comment parameter
    char* str = (char*) par;
    str_trunc( str );
    parameter.push_back( str );

  }

}

int Par::check_value( const char* str ) const {
  try {
    string tmp( str );
    validate( tmp );
    between_limits( tmp );
    return EXIT_SUCCESS;
  } catch( ParFileException& pfe ) {
    cerr << pfe;
    return EXIT_FAILURE;
  } catch( Exception& e ) {
    cerr << e;
    return EXIT_FAILURE;
  }

}

bool Par::is_indirrect( const string& str, char delimit ) {

  if ( delimit == str[ 0 ] )
    return true;
  else
    return false;

}

// Either min <= val or val <= max is not met
void Par::not_between_limits( char str[], 
			      const char* left, const string& left_val,
			      const char* right, const string& right_val )
  const {

  const char* format = "Par::not_between_limits( ) : "
    "For variable `%s',  %s (%s) >= %s (%s)\n";

  sprintf( str, format, parameter [ PARNAME ].c_str( ),
	   left, left_val.c_str( ), 
	   right, right_val.c_str( ) );

  return;

}

//
// Output
//    name = val  prompt
// to ostream os
//
void Par::plist( ostream& os ) const {

  const char* plist_format = "%15s = %-15s %s";
  const char* hidden_plist_format = "(%15s = %-15s) %s";

  if ( parameter[ PARMODE ] == "H" )
    return;
  
  const char* format = NULL;
  if ( parameter[ PARMODE ] == "h" )
    format = hidden_plist_format;
  else
    format = plist_format;

  size_t length = strlen( format ) + parameter[ PARNAME ].size() + 
    parameter[ PARVALUE ].size() + parameter[ PARPROMPT ].size() + 256;

  auto_ptr< char > mystr( new char[ length ] );

  char* ptr = &(*mystr);
    
  sprintf( ptr, format, parameter[ PARNAME ].data( ),
	     parameter[ PARVALUE ].data( ), parameter[ PARPROMPT ].c_str( ) );
  
  os << ptr;

}

static char mymsg[ 256 ];

void Par::pget_wrong_type( char msg[], const char* strtype ) const {

  const char* format = "Par::pgetb( ) : Parameter `%s' is not a %s type\n";

  if ( NUMTOKENS == parameter.size( ) )
    sprintf( msg, format, parameter[ PARNAME ].c_str( ), strtype );
  else
    sprintf( msg, format, "", strtype );

}

bool Par::pgetb( void ) const  {

  pget_wrong_type( mymsg, "boolean" );
  throw ParFileException( mymsg );

}

double Par::pgetd( void ) const  {

  pget_wrong_type( mymsg, "double" );
  throw ParFileException( mymsg );

}

int Par::pgeti( void ) const  {

  pget_wrong_type( mymsg, "int" );
  throw ParFileException( mymsg );

}

long Par::pgetl( void ) const  {

  pget_wrong_type( mymsg, "long" );
  throw ParFileException( mymsg );

}

void Par::pgetstr( char result[], size_t size ) const
   {

  pget_wrong_type( mymsg, "string" );
  throw ParFileException( mymsg );

}

string Par::pgetstring( ) const  {

  pget_wrong_type( mymsg, "string" );
  throw ParFileException( mymsg );

}

void Par::print( ostream& os ) const {

  vector< string >::const_iterator iter_begin =
    parameter.begin( );

  vector< string >::const_iterator iter_end =
    parameter.end( );

  for ( vector< string >::const_iterator iter = iter_begin;
	iter != iter_end; iter++ ) {
    os << *iter;
    vector< string >::const_iterator tmp = iter;
    if ( ++tmp != iter_end )
      os << char( DELIMIT );
  }

}

char* Par::save_string( char* str )  {

  if ( NULL == str || '\0' == *str )
    return NULL;

  try {

    char* new_str = new char[ strlen( str ) + 1 ];
    strcpy( new_str, str );
    return new_str;

  } catch ( bad_alloc& ba ) {

    string msg( "Par::save_string( " );
    msg.append( str );
    msg.append( " ) : Unable to allocate memory\n" );
    throw Exception( msg );

  }

}

int Par::count_tokens( char* str, char* delimit )  {

  try {

    int ntoken = 0;
    char* new_str = save_string( str );

    {
      char* token = strtok( new_str, delimit );
      while( NULL != token ) {
	ntoken++;
	token = strtok( NULL, delimit );
      }
    }

    delete [] new_str;

    return ntoken;

  } catch( Exception& e ) {
    throw;
  }

}

//
// A simple parser to tokenize a string.
// The tokens are stored in tokens, ntoken contains the number of tokens
//
int Par::my_tokenize( char *str, char *delimit, char ***tokens )
   {

    try {

      char* new_str = save_string( str );

      int ntoken = count_tokens( new_str, delimit );

      *tokens = new char* [ ntoken + 1 ];

      int num = 0;

      char* token = strtok( new_str, delimit );
      while( NULL != token ) {
	char* tmp = save_string( token );
	(*tokens)[ num++ ] = tmp;
	token = strtok( NULL, delimit );
      }
      (*tokens)[ num ] = (char *) NULL;   

      delete [] new_str;

      return num;

    } catch( Exception& e ) {
      throw e;
    }

}

void Par::delete_tokens( char** ptr ) {

  if ( ptr ) {
    int i=0;
    while( ptr[ i ] )  {
      delete [] ptr[ i ];
      i++;
    }
    delete [] ptr;
  }

  return;

}

///
void Par::print_tokens( char* str, char** str_argv, ostream& os ) {

  int i = 0;
  os << "The tokens for string `" << str << "' are :\n";

  while( str_argv[ i ] ) {
    os << "token[ " << i << " ] = `" << str_argv[ i ] << "'\n";
    i++;
  }

  return;

}

void Par::set_val( const string& str )  {

  throw ParFileException( "Par::set_val( " + str + " ) : Define me!\n" );

}

void Par::between_limits( const string& str ) const
   {

    throw ParFileException( "Par::between_limits( " + str +
			    " ) : Define me!\n" );

}

void Par::validate( const string& str ) const  {

    throw ParFileException( "Par::validate( " + str +
			    " ) : Define me!\n" );

}
