// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of paramxx
//
// paramxx is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// paramxx is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <errno.h>
#include <stdlib.h>
#include <string>
#include <fstream>

#include <getline/getline.h>
#include <suplib/str.h>
#include <suplibxx/io.h>

#include "ParFile.h"
#include "ParTxt.h"
#include "NameAttributeValue.h"
#include "BoolPar.h"
#include "CommentPar.h"
#include "LongPar.h"
#include "RealPar.h"
#include "StringPar.h"

using namespace std;

#define  MODE "mode"

ParFile::~ParFile( ) {
  if ( parameters ) {
    size_t count = 0;
    while( parameters[ count ] ) {
      delete parameters[ count ];
      ++count;
    }
    delete [] parameters;
    parameters = NULL;
  }
}

ParFile::ParFile( int argc, char** argv, const char* file ) : parameters( NULL ) {

    try {

      init_ParFile( argc, argv, file );

    } catch( ParFileException& pfe ) {
      throw;
    }

}

void ParFile::pget( const char* name, bool& result ) const{

    try {
      result = pgetb( name );
    } catch ( ParFileException& pfe ) {
      throw;
    }

}

void ParFile::pget( const char* name, double& result ) const {

    try {
      result = pgetd( name );
    } catch ( ParFileException& pfe ) {
      throw;
    }

}

void ParFile::pget( const char* name, int& result ) const {

    try {
      result = pgeti( name );
    } catch ( ParFileException& pfe ) {
      throw;
    }

}

void ParFile::pget( const char* name, long& result ) const {

    try {
      result = pgetl( name );
    } catch ( ParFileException& pfe ) {
      throw;
    }

}

void ParFile::pget( const char* name, char result[], size_t size ) const {

    try {
      pgetstr( name, result, size );
    } catch ( ParFileException& pfe ) {
      throw;
    }
    
}

void ParFile::pget( const char* name, string& result ) const {

    try {
      result = pgetstring( name );
    } catch ( ParFileException& pfe ) {
      throw;
    }
    
}

bool ParFile::pgetb( const char* name ) const  {

  try {
    Par* par = get_par( name );
    return par->pgetb( );
  } catch ( ParFileException& pfe ) {
    throw;
  }

}

double ParFile::pgetd( const char* name ) const {

  try {
    Par* par = get_par( name );
    return par->pgetd( );
  } catch ( ParFileException& pfe ) {
    throw;
  }

}

int ParFile::pgeti( const char* name ) const {

  try {
    Par* par = get_par( name );
    return par->pgeti( );
  } catch ( ParFileException& pfe ) {
    throw;
  }

}

long ParFile::pgetl( const char* name ) const {

  try {
    Par* par = get_par( name );
    return par->pgetl( );
  } catch ( ParFileException& pfe ) {
    throw;
  }

}

void ParFile::pgetstr( const char* name, char result[], size_t size ) const {

    try {
      Par* par = get_par( name );
      string str = par->pgetstring( );
      strncpy( result, str.c_str( ), size );
      result[ size - 1 ] = '\0';
      return;
    } catch ( ParFileException& pfe ) {
      throw;
    }

}

string ParFile::pgetstring( const char* name ) const {

    try {
      Par* par = get_par( name );
      return par->pgetstring( );
    } catch ( ParFileException& pfe ) {
      throw;
    }

}

// check to see the parameter mode of the file.
// The value of the parameter `mode' is the mode of the parameter file.
// If it is batch mode, then die on error, i.e. never ask any questions
// else querry the user till the right answer is enterred
void ParFile::get_mode_of_parameter_file( int argc, char** argv )  {

    try {

      // just in case the parameter MODE is not present in the parameter file.
      Par* par = search_name( MODE );

    } catch( ParFileException& pfe ) {

      return;

    }

    try {

      Par* par = search_name( MODE );

      // User did supplied a mode parameter in the parameter file.  Find out 
      // what it is then decide whether the program is to return or querry.
      set_return_on_error( par->get_mode( ) );

      // check if the user has entered the mode option at the command line
      for ( int i = 1; i < argc; i++ ) {

	NameAttributeValue cmdline_mode( argv[ i ] );

	if ( cmdline_mode.get_name( ) == MODE )
	  set_return_on_error( cmdline_mode.get_value( ) );

      }

    } catch( ParFileException& pfe ) {

      throw;

    }

  return;

}

Par* ParFile::get_par( const char* str ) const {

  try {

    const char* original_name = str;

    // indirrection level cannot exceed the
    // number of parameters within the file
    size_t max_count = 0;
    while( parameters[ max_count ] )
      ++max_count;

    string new_value;

    for ( size_t ii = 0; ii < max_count; ii++ ) {

      Par* par = search_name( str );

      new_value = par->get_value( );

      if ( Par::is_indirrect( new_value ) ) {
	str = new_value.c_str( );
	++str;
      } else
	return par;

    }

    // Have infinite loop
    const char* format =
      "ParFile::get_par( %s ) : Encountered an infinite loop\n";
    char tmp[ 256 ];
    sprintf( tmp, format, original_name );
    throw ParFileException( tmp );

  } catch ( ParFileException& pfe ) {
    throw;
  }

}

void ParFile::init_ParFile( int argc, char** argv, const char* file ) {

    try {

      // Set the default to return as though one is in a batch mode.
      return_on_error = true;

      read_ParFile( argc, argv, file );

      // All the parameters have been read.  Check to see what the mode of the
      // parameter file is.  From  this value, see if the program is to return
      // to user or querry the user untill a legitimate input is entered.
      get_mode_of_parameter_file( argc, argv );

      is_duplicate_cmdline_args( argc, argv );

      for ( int ii = 1; ii < argc; ii++ )
	setpar( argv[ ii ] );

    } catch( ParFileException& pfe ) {

      throw;

    } catch( Exception& e ) {

      throw;

    }

}

bool ParFile::is_parlist( string& cmdline_name, string& cmdline_value ) {

  if ( parameters ) {
    for ( size_t count = 0; NULL != parameters[ count ]; count++ )
      if ( parameters[ count ]->get_name( ) == cmdline_name )
	return false;
  }

  if ( 0 == cmdline_value.compare( "" ) )
    return true;
  else
    return false;

}

bool ParFile::is_duplicate_cmdline_args( int argc, char** argv ){

    vector< string > cmdline_options;

    // Store the names of command line arguments in the map
    for ( int i = 1; i < argc; i++ ) {

      NameAttributeValue arg( argv[ i ] );
      string name = arg.get_name( );
    
      vector< string >::const_iterator iter_begin = cmdline_options.begin( );
      vector< string >::const_iterator iter_end = cmdline_options.end( );
      for ( vector< string >::const_iterator iter = iter_begin;
	    iter != iter_end; iter++ )
	if ( *iter == name ) {
	  string msg( "ParFile::is_duplicate_cmdline_args( ) : Duplicate " );
	  msg += name;
	  throw ParFileException( msg );
	}

    }

    return false;
    
}

void ParFile::print( ostream& os ) const {
 
  if ( parameters ) {
    size_t count = 0;
    while( parameters[ count ] ) {
      parameters[ count ]->print( os );
      if ( parameters[ ++count ] )
	os << '\n';
    }
  }
  
}

void ParFile::querry_user( Par* item, char* str ) {

  if ( return_on_error )
    return;

  const char* format = "%s (%s) : ";
  
  string prompt = item->get_prompt();

  // Put in the initial value so the user can recall it if s/he wishes
  gl_histadd( str );

  do {

    size_t length =
      strlen( format ) + prompt.size( ) + strlen( str ) + 10;
    char* ptr = new char[ length ];

    sprintf( ptr, format, prompt.c_str( ), str );

    str = gl_getline( ptr );
    gl_histadd( str );
    delete [] ptr;
    ptr = NULL;

    if ( ! is_comment( str ) )
      str_prune( str );

  } while( EXIT_FAILURE == item->check_value( str ) );

  item->set_val( str );

  return;

}

void ParFile::set_return_on_error( const string& mode_of_parfile ) {

  if ( string::npos == mode_of_parfile.find( "b" ) )
    // user wants to be querried for the correct input
    return_on_error = false;
  else
    // The calling program is in a batch mode.  Do not querry for answer.
    return_on_error = true;

}

static char mytmp[ 512 ];

void ParFile::read_ParFile( int argc, char** argv, const char* file ) {

    try {

      parFilename.init_ParFilename( argc, argv, file );
      string fname( parFilename.get_filename( ) );

      ifstream input( fname.c_str( ), ios::in );
      if ( ! input ) {
	string msg( "ParFile::read_Parfile( ) : Unable to open file '" +
		    fname +"'\n" );
	throw ParFileException( msg );
      }

      parameters = new Par* [ 512 ];
      size_t count = 0;

      string mytxt;
      while( suplib::getrecord( input, mytxt, suplib::READ_LOGICAL |
				suplib::STRIP | suplib::CLEAN ) ) {

	ParTxt parTxt( (char*) mytxt.c_str( ) );
	char** buffer_argv = (char**) parTxt;

	if ( NULL == buffer_argv ) {

	  parameters[ count ] = new CommentPar( parTxt );

	} else if ( Par::BOOLEAN == buffer_argv[ Par::PARTYPE ][ 0 ] ) {

	  parameters[ count ] = new BoolPar( parTxt );

	} else if ( Par::INTEGER == buffer_argv[ Par::PARTYPE ][ 0 ] ) {

	  parameters[ count ] = new LongPar( parTxt );

	} else if ( Par::REAL == buffer_argv[ Par::PARTYPE ][ 0 ] ) {

	  parameters[ count ] = new RealPar( parTxt );

	} else if ( Par::STRING == buffer_argv[ Par::PARTYPE ][ 0 ] ) {

	  parameters[ count ] = new StringPar( parTxt );

	}

	count++;

	// delete [] txt;

      }

      parameters[ count ] = NULL;

    } catch( ParFileException& pfe ) {
      throw;
    }

}

Par* ParFile::search_name( const char* str ) const {

    if ( parameters ) {
      for ( size_t count = 0; NULL != parameters[ count ]; count++ )
	if ( parameters[ count ]->get_name( ) == str )
	  return parameters[ count ];
    }
    
    string msg( "ParFile::search_name( " );
    msg.append( str );
    msg.append( " ) : Unable to find the parameter\n" );
  
    throw ParFileException( msg );
  
}


void ParFile::setpar( char* str ) {

  try {

    NameAttributeValue cmdline( str );

    string cmdline_name( cmdline.get_name( ) );

    // If the user has entered PFILE=foo.par at the command line then the 
    // file `foo.par' is to be used as the parameter file, if it exists.
    // Hence do not set the parameter to anything.  Simply return.
    if ( cmdline_name == PFILE )
      return;

    string cmdline_value( cmdline.get_value( ) );

    /*
    cout << "ParFile::setpar( " << str << " ) : "
	 << "name = `" << cmdline.get_name( ) << '\t';
    cout << "value = `" << cmdline.get_value( ) << "'\n";
    */

    // parlist foo.par
    if ( is_parlist( cmdline_name, cmdline_value ) )
      return;

    Par* par = search_name( cmdline_name.c_str( ) );

    if ( return_on_error )
      // In batch mode, ie the parameter mode contains a b
      par->set_val( cmdline_value.c_str( ) );
    else {
      // User has requested to be querried, ie the parameter mode contains a q
      if ( EXIT_FAILURE == par->check_value( cmdline_value.c_str( ) ) ) {
	querry_user( par, (char*) cmdline_value.c_str( ) );
      } else {
	par->set_val( cmdline_value.c_str( ) );
      }
    }

  } catch( ParFileException& pfe ) {

    throw;

  } catch( Exception& e ) {

    throw;

  }

}

int ParFile::traverse( int (*fct)( Par* ) ) const {

  if ( parameters ) {
    size_t count = 0;
    while( parameters[ count ] ) {
      int status = (*fct)( parameters[ count++ ] );
      if ( 0 != status )
	return status;
    }
  }

  return 0;

}

int ParFile::is_blank( const char *str ) {

  while( isspace( *str++ ) )
    ;

  return ! (*--str);

}

int ParFile::is_comment( const char* str, char delimit ) {

  // move to first non-space character
  while ( isspace( *str ) )
    str++;

  /**
   * if the first character is #, or the line is blank,
   * then the line is a comment 
   */
  return str[0] == delimit || ParFile::is_blank( str );

}

#undef MODE
