#ifndef PARFILE_H
#define PARFILE_H

// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of paramxx
//
// paramxx is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// paramxx is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <iostream>

#include "Par.h"
#include "ParFilename.h"

using namespace std;

/**
 * To get the parameters in the parameter file.
 */
class ParFile {

  /**
   * This operator outputs the contents of the object.
   */
  friend ostream& operator << ( ostream& os, ParFile& parFile ) {
    parFile.print( os );
    return os;
  }

  /**
   * This operator outputs the contents of the object.
   */
  friend ostream& operator << ( ostream& os, ParFile* parFile ) {
    parFile->print( os );
    return os;
  }

public:

  /**
   * The destructor.
   */
  ~ParFile( );

  /**
   * The default constructor.
   */
  ParFile( ) : parameters( NULL ) { }

  /**
   * The constructor to load the parameter file and parse the command line
   * args. Calling this constructor is equivalent to call the default 
   * constructor then call the init_ParFile method.
   * \param argc The number of parameters from the command line.
   * \param argv The command line arguments.
   * \param file The specific file to use.
   * \return void
   * \throw  ParFileException
   */
  ParFile( int argc, char** argv, const char* file=NULL );

  /**
   * Get the parameter filename.
   * \return The name of the parameter file.
   */
  string get_filename( ) { return parFilename.get_filename( ); }

  /**
   * Load the appropiate parameter file and parse the command line args.
   * \param argc The number of parameters from the command line.
   * \param argv The command line arguments.
   * \param file The specific file to use.
   * \return void
   * \throw  ParFileException
   */
  void init_ParFile( int argc, char** argv, const char* file=NULL );

  /// Get the pointer to the parameter
  Par* get_par( const char* name ) const;

  /**
   * Loop over the parameters to call the print method for each parameter.
   * The individual parameter will print its content in the form:
   * name,type,mode,val,min,max,prompt
   * \param os The stream to print the content of the parameter file.
   */ 
  void print( ostream& os ) const;

  /**
   * To get a parameter of type boolean.
   * \param  name The name of the parameter to retrieve its value
   * \return The value of the parameter name.
   * \throw  ParFileException
   */
  bool pgetb( const char* name ) const;

  /**
   * To get a parameter of type double.
   * \param  name The name of the parameter to retrieve its value
   * \return The value of the parameter name.
   * \throw  ParFileException
   */
  double pgetd( const char* name ) const;

  /**
   * To get a parameter of type long.
   * \param  name The name of the parameter to retrieve its value
   * \return The value of the parameter name.
   * \throw  ParFileException
   */
  int pgeti( const char* name ) const;

  /**
   * To get a parameter of type long.
   * \param  name The name of the parameter to retrieve its value
   * \return The value of the parameter name.
   * \throw  ParFileException
   */
  long pgetl( const char* name ) const;

  /**
   * To get a parameter of type char[].
   * The length of the result in the following function must be at least size+1
   * \param  name The name of the parameter to retrieve its value
   * \return The value of the parameter name.
   * \throw  ParFileException
   */
  void pgetstr( const char* name, char result[], size_t size ) const;

  /**
   * To get a parameter of type string.
   * \param  name The name of the parameter to retrieve its value
   * \return The value of the parameter name.
   * \throw  ParFileException
   */
  string pgetstring( const char* name ) const;

  /// Same functionality as ParFile::pgetb
  void pget( const char* name, bool& result ) const;

  /// Same functionality as ParFile::pgetd
  void pget( const char* name, double& result ) const;

  /// Same functionality as ParFile::pgeti
  void pget( const char* name, int& result ) const;

  /// Same functionality as ParFile::pgetl
  void pget( const char* name, long& result ) const;

  /// Same functionality as ParFile::pgetstr
  void pget( const char* name, char result[], size_t ) const;

  /// Same functionality as ParFile::pgetstring
  void pget( const char* name, string& rult ) const;

  /**
   * Traverse the parameters list and invoke the function fct, if fct
   * returns a non-zero the traversal is aborted and the fct return value
   * is returned.
   */
  int traverse( int (*fct)( Par* ) ) const;

  static int is_blank( const char* str );

  static int is_comment( const char* str, char delimit='#' );

private:

  // Whether or not to querry the user, useful for interactive or batch mode.
  bool          return_on_error;

  // To look for the parameter file
  ParFilename parFilename;

  // To hold the parameters in the paramter file parFilename.
  Par** parameters;

  void get_mode_of_parameter_file( int argc, char** argv );

  bool is_duplicate_cmdline_args( int argc, char** argv );

  bool is_parlist( string& cmdline_name, string& cmdline_value );

  void querry_user( Par* par, char* str );

  /**
   * Read the parameter file.
   */
  void read_ParFile( int argc, char** argv, const char* file );

  void setpar( char* );

  void set_return_on_error( const string& str );

  /**
   * Search for the parameter str.
   */
  Par* search_name( const char* str ) const;

  /**
   * Declare but purposely do not define
   */
  ParFile( const ParFile& pf );

  /**
   * Declare but purposely do not define
   */
  ParFile& operator = ( const ParFile& pf );

};

/**
 * \example ptest.cc
 * The classic way to read the parameter file.
 */

/**
 * \example ptest.par
 * Given a parameter file of the form: 
 */

#endif
